package dev.pfaff.altargui.pathfinder.fontkit;

import dev.pfaff.altargui.libimage.ColorType;
import dev.pfaff.altargui.libimage.ImageData;
import dev.pfaff.altargui.libimage.ImageInfo;
import dev.pfaff.altargui.texture.LoadableTexture;
import dev.pfaff.altargui.types.Color;
import dev.pfaff.altargui.util.GLThread;

import java.io.IOException;
import java.lang.foreign.Arena;
import java.lang.foreign.ValueLayout;
import java.nio.ByteOrder;

import static dev.pfaff.altargui.util.BasicFuture.executeOn;
import static java.lang.foreign.ValueLayout.JAVA_INT;

public record FKGlyphLoadableTexture(FKFontRef font, int codepoint, Factors factors) implements LoadableTexture {
	private static final ValueLayout.OfInt JAVA_INT_BIG_ENDIAN = ValueLayout.JAVA_INT.withOrder(
			ByteOrder.BIG_ENDIAN);

	private static final int SIZE = 64;
	public static final FKCanvasRasterizationOptions RASTERIZATION_OPTIONS = FKCanvasRasterizationOptions.SubpixelAa;

	public static Factors defaultFactors = new Factors(0.0f, 0.0f, 0.0f, 0.0f);

	public record Factors(float originX, float originY, float width, float height) {
	}

	public FKGlyphLoadableTexture(FKFontRef font, int codepoint) {
		this(font, codepoint, defaultFactors);
	}

	@Override
	public ImageData load() throws IOException {
		// inspection is misguided
		//noinspection DataFlowIssue
		var idOrNull = executeOn(() -> this.font.glyphForCodepoint(this.codepoint), GLThread::submitToGLThread);
		// inspection is wrong
		//noinspection ConstantValue
		if (idOrNull == null) throw new IOException("The font does not contain the specified codepoint");
		int id = idOrNull;
		var buf = Arena.ofAuto().allocate(JAVA_INT, SIZE * SIZE);
		var rasterBounds = executeOn(() -> this.font.rasterBounds(id, SIZE, RASTERIZATION_OPTIONS), GLThread::submitToGLThread);
		var translationX = rasterBounds.originX() * this.factors.originX + rasterBounds.width() * this.factors.width;
//		var translationY = rasterBounds.originY() * this.factors.originY + rasterBounds.height() * this.factors.height;
		var translationY = SIZE - (rasterBounds.originY() * this.factors.originY + rasterBounds.height() * this.factors.height);
		executeOn(() -> this.font.rasterizeCodepoint(buf,
				SIZE,
				SIZE,
				FKCanvasFormat.Rgba32,
				id,
				SIZE,
				translationX,
				translationY,
				RASTERIZATION_OPTIONS), GLThread::submitToGLThread);
		for (long i = 0; i < buf.byteSize(); i += JAVA_INT_BIG_ENDIAN.byteSize()) {
			buf.set(JAVA_INT_BIG_ENDIAN, i, Color.fromRgba(buf.get(JAVA_INT_BIG_ENDIAN, i)).unpremultiply().rgba());
		}
		return new ImageData(new ImageInfo(SIZE, SIZE, ColorType.Rgba8), buf);
	}
}
